<?php

namespace Intergo\OptOut;

use Intergo\OptOut\Enums\OptOutEnum;
use Intergo\OptOut\Repositories\ContactRepository;
use Illuminate\Redis\Connections\Connection;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Redis;
use Intergo\Helpers\HashHelper;

abstract class AbstractOptOutService
{
    protected Redis $redis;
    protected DB $db;

    /**
     * AbstractOptOutService constructor.
     */
    public function __construct(Redis $redis, DB $db)
    {
        $this->redis = $redis;
        $this->db = $db;
    }

    /**
     * @return string
     */
    protected function cacheConnectionName()
    {
        return OptOutEnum::CACHE_CONNECTION;
    }

    /**
     * @return Connection
     */
    private function cacheConnection(): Connection
    {
        return $this->redis::connection(self::cacheConnectionName());
    }

    /**
     * @param OptOutVO $optOutVO
     * @return void
     */
    public function cacheOptOuts(OptOutVO $optOutVO)
    {
        $this->cacheGlobalOptOuts($optOutVO->authUserId, $optOutVO->appUserId);
        if (empty($optOutVO->listIds)) {
            return;
        }
        $this->cacheListOptOuts($optOutVO->authUserId, $optOutVO->listIds);
    }

    /**
     * @param $userId
     * @param $appUserId
     * @return void
     */
    public function cacheGlobalOptOuts($userId, $appUserId)
    {
        $key = $this->key($userId, OptOutEnum::GLOBAL_OPT_OUT);
        if ($this->cacheConnection()->exists($key)) {
            return;
        }
        $optOuts = (new ContactRepository($this->db))->getOptOutsByUserId($appUserId);
        //Cache result to avoid multiple query executions
        $this->setOnCache($key, empty($optOuts));
        foreach ($optOuts as $optOut) {
            $this->setOnCache($this->key($userId, $optOut->phone));
        }
    }

    /**
     * @param $userId
     * @param $listIds
     * @return void
     */
    public function cacheListOptOuts($userId, $listIds)
    {
        $key = $this->key($userId, OptOutEnum::LIST_OPT_OUT);
        if ($this->cacheConnection()->exists($key)) {
            return;
        }
        [$listIds, $listKey] = $this->processListIdsAndGetListKey($listIds);
        $optOuts = (new ContactRepository($this->db))->getOptOutsByListIds($listIds);
        //Cache result to avoid multiple query executions
        $this->setOnCache($key, empty($optOuts));
        foreach ($optOuts as $optOut) {
            $this->setOnCache($this->getListOptOutKey($userId, $listKey, $optOut->phone));
        }
    }

    /**
     * Check if a number is unsubscribed
     *
     * @param $userId
     * @param $phoneNumber
     * @param $listIds
     * @param ?bool $bypassOptOut
     * @return array|false|mixed
     */
    public function isUnsubscribed($userId, $phoneNumber, $listIds = null, ?bool $bypassOptOut = false)
    {
        if ($bypassOptOut) {
            return false;
        }
        if ($isGlobalOptOut = $this->isGlobalOptOut($userId, $phoneNumber)) {
            return $isGlobalOptOut;
        }
        if (empty($listIds)) {
            return false;
        }
        [$listIds, $listKey] = $this->processListIdsAndGetListKey($listIds);
        return $this->cacheConnection()->get($this->getListOptOutKey($userId, $listKey, $phoneNumber)) ?? false;
    }

    /**
     * @param $userId
     * @param $phoneNumber
     * @return boolean
     */
    public function isGlobalOptOut($userId, $phoneNumber)
    {
        return $this->cacheConnection()->get($this->key($userId, $phoneNumber)) ?? false;
    }

    /**
     * @param $userId
     * @param $listKey
     * @param $phoneNumber
     * @return string
     */
    public function getListOptOutKey($userId, $listKey, $phoneNumber): string
    {
        return $this->key($userId, "$listKey:$phoneNumber");
    }

    /**
     * @param $listIds
     * @return array
     */
    public function processListIdsAndGetListKey($listIds): array
    {
        $listIds = is_array($listIds) ? $listIds : [$listIds];
        sort($listIds);
        $listKey = HashHelper::hash(implode(':', $listIds));
        return [$listIds, $listKey];
    }

    /**
     * @param $authUserId
     * @return void
     */
    public function invalidateOptOutFlags($authUserId)
    {
        $keyGlobal = $this->key($authUserId, OptOutEnum::GLOBAL_OPT_OUT);
        $keyLists = $this->key($authUserId, OptOutEnum::LIST_OPT_OUT);
        $this->cacheConnection()->del($keyGlobal, $keyLists);
    }

    /**
     * @param $userId
     * @param string $type
     * @return string
     */
    public function key($userId, string $type): string
    {
        return OptOutEnum::cacheKey($userId, $type);
    }

    /**
     * @param string $key
     * @param bool $value
     * @return void
     */
    private function setOnCache(string $key, bool $value = true): void
    {
        $this->cacheConnection()->set($key, $value,
            OptOutEnum::OPT_OUT_CACHE_RESOLUTION,
            OptOutEnum::OPT_OUT_CACHE_TIME);
    }
}
